"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const path = tslib_1.__importStar(require("path"));
const fs = tslib_1.__importStar(require("fs-extra"));
const format_1 = require("../tools/format");
const now = new Date();
const timeString = `${now.getFullYear()}-${now.getMonth() +
    1}-${now.getDate()} ${now.getHours()}:${now.getMinutes()}:${now.getSeconds()}`;
/* istanbul ignore next */
async function print(entity, options, includeMap) {
    if (!entity || !entity.ns) {
        // 过滤掉没有 namespace 的
        return;
    }
    const ns = options.autoNamespace
        ? printNamespace(entity, options.root)
        : entity.ns;
    const content = `// prettier-ignore
// generate${options.useTimestamp ? ` at ${timeString}` : ''} by dts-from-thrift
declare namespace ${ns} {
${printEnums(entity)}
${printInterfaces(entity)}
${printTypeDefs(entity)}
${printServices(entity.services)}
}
`;
    const relativePath = path.parse(path.relative(options.root, entity.fileName));
    const target = path.resolve(options.tsRoot, relativePath.dir, `${relativePath.name}.d.ts`);
    await fs.ensureFile(target);
    return fs.writeFile(target, format_1.prettier(fixIncludeNamespace(content, entity, includeMap)), 'utf8');
}
exports.print = print;
/**
 * 根据到 root 的相对路径生成 namespace
 *
 * @param {RpcEntity} entity
 * @param {string} root
 * @returns {string} namespace
 */
function printNamespace(entity, root) {
    const relativePath = path.parse(path.relative(root, entity.fileName));
    const ns = relativePath.dir.replace(new RegExp(`\\${path.sep}`, 'g'), '.');
    /* istanbul ignore if */
    if (ns.indexOf('.') === 0) {
        throw new Error('ns error');
    }
    return ns;
}
/**
 * 输出 enums
 *
 * @export
 * @param {RpcEntity} entity thrift entity
 * @returns {string}
 */
function printEnums(entity) {
    /* istanbul ignore if */
    if (!entity.enums || entity.enums.length === 0) {
        return '';
    }
    let rtn = '';
    entity.enums.forEach((datum, _i) => {
        rtn += `  export const enum ${datum.name} {
    ${Object.keys(datum.properties)
            .map((key, i, arr) => {
            const property = datum.properties[key];
            return attachComment(`${key} = ${property.value}${arr.length - 1 === i ? '' : ','}`, property.comment);
        })
            .join('\n    ')}
  }

`;
    });
    return rtn;
}
exports.printEnums = printEnums;
/**
 * 输出 interfaces
 *
 * @export
 * @param {Pick<RpcEntity, 'interfaces'>} entity thrift entity
 * @returns {string}
 */
function printInterfaces(entity) {
    if (!entity.interfaces || entity.interfaces.length === 0) {
        return '';
    }
    let rtn = '';
    entity.interfaces.forEach((datum, _i) => {
        rtn += `  export interface ${datum.name} {
    ${Object.keys(datum.properties)
            .map((key, i, arr) => {
            const property = datum.properties[key];
            return attachComment(`${key}${property.optional ? '?' : ''}: ${property.type};`, property.comment, property.defaultValue);
        })
            .join('\n    ')}
  }

`;
    });
    return rtn;
}
exports.printInterfaces = printInterfaces;
/**
 * 输出 typedef
 *
 * @export
 * @param {RpcEntity} entity
 * @returns {string}
 */
function printTypeDefs(entity) {
    if (entity.typeDefs.length === 0) {
        return '';
    }
    let rtn = '';
    entity.typeDefs.forEach(datum => {
        rtn += `  export type ${datum.alias} = ${datum.type};\n`;
    });
    return rtn + '\n';
}
exports.printTypeDefs = printTypeDefs;
/**
 * 以 12 个半角字符宽度对齐行注释
 *
 * @param {string} str
 * @param {string} comment
 * @returns {string}
 */
function attachComment(str, comment, defaultValue) {
    let c = comment;
    if (defaultValue) {
        if (!c) {
            c = `(default: ${defaultValue})`;
        }
        else {
            c += ` (default: ${defaultValue})`;
        }
    }
    if (!c) {
        return str;
    }
    const len = str.length + 1;
    const len2 = Math.ceil(len / 12) * 12;
    return `${str.padEnd(len2, ' ')}// ${c}`;
}
/**
 * 这个函数用来修复类型是通过 import 导入的名称
 *
 * @export
 * @param {string} content
 * @param {RpcEntity} entity
 * @param {{ [key: string]: RpcEntity }} includeMap
 * @returns {string}
 */
function fixIncludeNamespace(content, entity, includeMap) {
    let result = content;
    const currentFile = path.parse(entity.fileName);
    entity.includes.forEach(include => {
        // 获取所有 include 语句所对应的 entity
        const includeKey = path.resolve(currentFile.dir, include);
        const includeEntity = includeMap[includeKey];
        if (!includeEntity) {
            return;
        }
        const shouldBeNS = includeEntity.ns; // 需要替换成的 namespace
        const shouldReplaceNS = path.parse(include).name; // 如果文件中有用到 filename 作为 namespace
        result = result.replace(new RegExp(`([^\\w])${shouldReplaceNS}\\.`, 'g'), `$1${shouldBeNS}.`);
    });
    return format_1.prettier(result);
}
exports.fixIncludeNamespace = fixIncludeNamespace;
function printServices(serviceEntity) {
    return serviceEntity.reduce((rtn, cur) => {
        rtn += `  export interface ${cur.name} {
${Object.keys(cur.interfaces)
            .map(key => {
            const i = cur.interfaces[key];
            return attachComment(`    ${key}(req: ${i.inputType}): Promise<${i.returnType}>;`, i.comment);
        })
            .join('\n')}
  }

`;
        return rtn;
    }, '');
}
exports.printServices = printServices;
