"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const fs = tslib_1.__importStar(require("fs-extra"));
const parseNamespace_1 = require("./parseNamespace");
const parseInclude_1 = require("./parseInclude");
const parseEnum_1 = require("./parseEnum");
const parseStruct_1 = require("./parseStruct");
const parseTypedef_1 = require("./parseTypedef");
const parseService_1 = require("./parseService");
async function readCode(filefullname, options, includeMap) {
    const codes = (await fs.readFile(filefullname, 'utf8'))
        .split(/(\r\n)|(\n)/g)
        .filter(d => !!d);
    // 预处理单行 { 的情况，合并到前面一行
    for (let i = codes.length - 1; i >= 0; i--) {
        if (/^\s*{\s*(#.*|\/\/.*|\/\*.*)?$/.test(codes[i].trim())) {
            codes[i - 1] += codes[i];
        }
    }
    const rtn = {
        ns: '',
        fileName: filefullname,
        includes: [],
        interfaces: [],
        enums: [],
        typeDefs: [],
        services: []
    };
    let currentLine = '';
    let i = 0;
    let blockStartType = ''; // block statement 类型
    let blockStartIndex = 0; // block statement 开始位置
    while ((currentLine = codes[i++])) {
        let tmp = { hit: false, mc: null };
        // 0. 首先判断是否是 block 结束
        if ((blockStartType === 'enum' ||
            blockStartType === 'struct' ||
            blockStartType === 'service') &&
            currentLine.trim().match(/^}/)) {
            const blockLines = codes.slice(blockStartIndex, i);
            switch (blockStartType) {
                case 'enum':
                    rtn.enums.push(parseEnum_1.parseEnum(blockLines, options));
                    break;
                case 'struct':
                    rtn.interfaces.push(parseStruct_1.parseStruct(blockLines, options));
                    break;
                case 'service':
                    rtn.services.push(parseService_1.parseService(blockLines, options));
                    break;
            }
            blockStartType = '';
            continue;
        }
        else if (blockStartType === 'comment') {
            if (currentLine.trim().match(/\*\//)) {
                blockStartType = '';
            }
            continue;
        }
        // 检查多行注释的开头
        if (isCommentBlockStart(currentLine)) {
            blockStartType = 'comment';
            blockStartIndex = i - 1;
            continue;
        }
        // 1. 检查是否是 namespace 语句
        // namespace 不同的情况没处理，需手动解决
        if (!rtn.ns) {
            tmp = parseNamespace_1.parseNamespace(currentLine, options);
            if (tmp.hit) {
                rtn.ns = tmp.mc;
                continue;
            }
        }
        // 2. 检查 include 语句
        tmp = parseInclude_1.parseInclude(currentLine, options);
        if (tmp.hit) {
            rtn.includes.push(tmp.mc);
            continue;
        }
        // 3. 检查 enum 语句
        if (isEnumStart(currentLine)) {
            blockStartType = 'enum';
            blockStartIndex = i - 1;
            continue;
        }
        // 4. 检查 struct 语句
        if (isStructStart(currentLine)) {
            blockStartType = 'struct';
            blockStartIndex = i - 1;
            continue;
        }
        tmp = parseService_1.isServiceBlockStart(currentLine);
        if (tmp.hit && tmp.mc) {
            blockStartType = 'service';
            blockStartIndex = i - 1;
            continue;
        }
        // 5. 检查 typedef 语句
        const defTemp = parseTypedef_1.parseTypedef(currentLine, options);
        if (defTemp) {
            rtn.typeDefs.push(defTemp);
            continue;
        }
    }
    /* istanbul ignore if */
    if (includeMap) {
        includeMap[rtn.fileName] = rtn;
    }
    return rtn;
}
exports.readCode = readCode;
function isStructStart(line) {
    return /^struct\s+([\w\d_]+)(?:\s+)?\{/.test(line.trim());
}
exports.isStructStart = isStructStart;
// function isCommentLine(line: string): boolean {
//   const tline = line.trim();
//   return tline.indexOf('//') === 0 || tline.indexOf('#') === 0;
// }
function isCommentBlockStart(line) {
    return line.trim().match(/^\s*\/\*/) !== null && line.indexOf('*/') === -1;
}
function isEnumStart(line) {
    return /^enum\s/.test(line.trim());
}
